import numpy as np

from njet.functions import cos
from lieops import create_coords, construct

# N.B. the length of an element will be used only later, when it comes to calculating the flow.

# Reference(s):
# [1] M. Titze: "Approach to Combined-function magnets via symplectic slicing", Phys. Rev. STAB 19 054002 (2016)
# [2] M. Titze: "Space Charge Modeling at the Integer Resonance for the CERN PS and SPS", PhD Thesis (2019)

class hard_edge_element:
    
    def __init__(self, *args, length=1, **kwargs):
        '''
        Class to model the Hamiltonian of a 6D hard-edge element.
        '''
        self.length = length
        if 'beta0' in kwargs.keys():
            self.calcHamiltonian(*args, **kwargs)
        if hasattr(self, 'full_hamiltonian'):
            self.setHamiltonian(**kwargs)
        
    def calcHamiltonian(self, beta0, sqrtexp: int=2, **kwargs):
        '''
        Compute the Hamiltonian of a drift.
        For the underlying coordinate system and further details see Refs. [1, 2] below.
        In particular, the Hamiltonian is given by Eq. (2.16) in Ref. [2].

        Parameters
        ----------
        sqrtexp or power: int, optional
            Power up to which the square root of the drift should be expanded. 
        '''
        assert 0 < beta0 and beta0 < 1
        kwargs['power'] = kwargs.get('power', sqrtexp)
        # Compute the CFM drift part
        x, y, sigma, px, py, psigma = create_coords(3, real=True, **kwargs)
        one_hateta2 = lambda ps: ((1 + ps*beta0**2)**2 - 1 + beta0**2)/beta0**2 # Eqs. (15c) and (17) in Ref. [1]
        sqrt = lambda p1, p2, ps: (one_hateta2(ps) - p1**2 - p2**2)**(1/2)
        drift_s = construct(sqrt, px, py, psigma, **kwargs) # expand sqrt around [px, py, psigma] = [0, 0, 0] up to power
        hamiltonian = psigma - drift_s
        hamiltonian.pop((0, 0, 0, 0, 0, 0), None)
        self.full_hamiltonian = hamiltonian
        self._prop = {}
        self._prop['beta0'] = beta0
        self._prop['sqrtexp'] = sqrtexp
        self._prop['drift'] = hamiltonian
        self._prop['drift_sqrt'] = drift_s
        self._prop['full'] = hamiltonian
        self._prop['coords'] = x, y, sigma, px, py, psigma
        
    def setHamiltonian(self, *projection, **kwargs):
        '''
        Set Hamiltonian by dropping components associated with terms not in given list.
        
        self.hamiltonian will be updated.
        
        !! Attention !!
        After this operation it may become necessary to re-caclulate flows and the one-turn map again.
        
        Parameters
        ----------
        projection: int
            The indices of the coordinates of interest. Must reach from 0 to self.full_hamiltonian.dim.
            If there are no arguments given, then the default Hamiltonian will be restored.
        '''
        # consistency checks
        new_dim = len(projection)
        ham = self.full_hamiltonian
        if new_dim == 0: # default: 6D Hamiltonian
            projection = range(ham.dim)
            new_dim = ham.dim
        assert new_dim <= ham.dim and max(projection) < ham.dim, 'Number of spatial dimensions too large.'
        projection = list(projection) + [p + ham.dim for p in projection] # the eta-components duplicate the indices.
        complement = [k for k in range(2*ham.dim) if not k in projection]
        new_values = {}
        for k, v in ham.items():
            if any([k[p] != 0 for p in complement]): # only keep those coefficients which do not couple to other directions
                continue
            new_values[tuple([k[p] for p in projection])] = v
        self.hamiltonian = ham.__class__(values=new_values, dim=new_dim, max_power=ham.max_power)
        
    def copy(self):
        result = self.__class__(length=self.length)
        # copy all the fields
        for field, value in self.__dict__.items():
            setattr(result, field, value)
        return result
        
class phaserot(hard_edge_element):
    def __init__(self, *tunes, length=1, **kwargs):
        '''
        A generic uncoupled phase (space) rotation.
        
        Parameters
        ----------
        tunes: float or array_like
            Tune(s) defining the phase rotation.
        '''
        self.tunes = tunes
        self.length = length
        if len(tunes) > 0:
            self.calcHamiltonian(*tunes, **kwargs)
        if hasattr(self, 'full_hamiltonian'):
            self.setHamiltonian(**kwargs)
        
    def calcHamiltonian(self, *tunes, **kwargs):
        dim = len(tunes)
        xieta = create_coords(dim=dim)
        self.full_hamiltonian = sum([-tunes[k]*xieta[k]*xieta[k + dim] for k in range(dim)])
        
        
class cfm(hard_edge_element):
    def __init__(self, components=[0], tilt=0, *args, **kwargs):
        '''
        Class to model a combined-function-magnetc (cfm).

        Parameters
        ----------
        beta0: float
            Relativistic beta = v/c of the reference trajectory.

        components: list, optional
            A list of complex entries, denoting the components of the cfm field.
            The components are hereby given with respect to a transversal axis, tilted against the Serre-Frenet
            coordinate system by an angle of choice (see tilt parameter below and Eq. (2.40) in Ref. [2]).

            If K_x = 0 this transverse axis smears out a cone. The ideal trajectory is hereby given by a horizontal
            cut of the cone with a plane.
            
            Note that additional zeros in the list will cause the routine to compute more orders of the CFM.
            
        tilt: float, optional
            The tilt between the dipole component and the higher-order magnet components of the cfm.
            
        **kwargs
            Optional arguments passed to self.calcHamiltonian and self.setHamiltonian.
        '''
        self.components = components
        self.tilt = tilt
        hard_edge_element.__init__(self, *args, **kwargs)
        
    def setHamiltonian(self, *args, **kwargs):
        '''
        Set self.hamiltonian to requested dimension.
        
        Parameters
        ----------
        style: str, optional
            Name of the key in self._prop denoting the Hamiltonian to be used.
            Supported options are:
            'full': use the full Hamiltonian (default)
            'kick': Only use the Hamiltonian containing field-components (see Eq. (2.36) in Ref. [2])
            'drift': Only the drift part of the Hamiltonian is used, i.e. all fields switched off.
        '''
        supported_styles = ['full', 'kick', 'drift']
        if not hasattr(self, '_style') or 'style' in kwargs.keys():
            self._style = kwargs.get('style', 'full')
        if not self._style in supported_styles:
            raise RuntimeError(f"Style '{self._style}' not recognized. Supported styles:\n{supported_styles}")
        self.full_hamiltonian = self._prop[self._style]
        hard_edge_element.setHamiltonian(self, *args, **kwargs)
    
    def _g(self):
        '''
        Compute the g-part of the cfm. Code taken in parts from my MAD-X implementation 
        (see https://github.com/mtitze/MAD-X/blob/master/src/trrun.f90).
        '''
        exp = lambda z: complex(np.exp(z)) # convert numpy.float64 back to float as precaution to prevent unpredictable results if using jets etc.

        # Eq. (2.40) resp. (2.47) in Ref. [2]:
        # (By + 1j*Bx)_(phi = phi0) = sum_k components[k] r^k,
        # where phi0 = tilt here.

        # since kx = real(components[0]) and ky = -imag(components[0]), i.e.
        # components[0] = kx - 1j*ky,
        # we have:
        kappa = self.components[0].conjugate()
        barkappa = self.components[0]

        nord = len(self.components) - 1
        # Now fill up the g_{ij}'s for j = 0, ..., i and i = 0, ..., nord + 1.
        g = {}
        g[(0, 0)] = 0
        g[(1, 0)] = -barkappa
        g[(1, 1)] = -kappa # = g[(1, 0)].conjugate()
        for k in range(1, nord + 1):
            for j in range(k):
                # Eq. (6), in Ref. [1]
                g[(k + 1, j + 1)] = ( barkappa*g[(k, j + 1)]*(j + 1)*(j - k + 3/2) + 
                                     kappa*g[(k, j)]*(k - j)*(1/2 - j) )/(k - j)/(j + 1)
            # Eq. (8) in Ref. [1]
            sum0 = 0
            for j in range(1, k + 1):
                sum0 = sum0 - (k + 1 - j)*g[(k + 1, j)]*exp(-self.tilt*2*1j*j)
            g[(k + 1, 0)] = ( sum0 - 2**k*exp(-self.tilt*1j*k)*( self.components[k] 
                            + 1/2*(barkappa*exp(self.tilt*1j) + kappa*exp(self.tilt*-1j))*self.components[k - 1] ) )/(k + 1)
            g[(k + 1, k + 1)] = g[(k + 1, 0)].conjugate()
        return g

    def calcHamiltonian(self, **kwargs):
        '''
        Compute the Hamiltonian of the cfm (without electric fields).
        For the underlying coordinate system and further details see Refs. [1, 2] below.
        In particular, the Hamiltonian is given by Eq. (2.16) in Ref. [2].

        Parameters
        ----------
        **kwargs
            Optional keyword arguments passed to 'construct' routine.

        Returns
        -------
        dict
            A dictionary having the following keys:
            
            kx: float
                The field strength of the x-component of the dipole.
                
            ky: float
                The field strength of the y-component of the dipole.

            H: liepoly
                The Hamiltonian of the cfm up to the requested order.

            H_drift: liepoly
                The drift part of H.

            H_field: liepoly
                The field part of H.
                
            g: dict
                The g-components in Ref. [1]
            G: dict
                The G-components in Ref. [1]

            So that H = H_drift + H_field.
        '''
        # compute the Hamiltonian of the drift
        hard_edge_element.calcHamiltonian(self, **kwargs)
        x, y, sigma, px, py, psigma = self._prop['coords']
        # Compute the CFM vector potential
        # G = (1 + Kx*x + Ky*y)*A_t near Eq. (2).
        # Here G := G*e/p0 (The factor e/p0 can be absorbed in the coefficients later; see also the comment on page 4, right column)
        g = self._g()
        rp = (x + y*1j)/2
        rm = rp.conjugate()
        G = sum([rp**(k - j)*rm**j*g[(k, j)] for k, j in g.keys()]) # N.B.: We need to multiply g[(k, j)] from right in case their entries are jetpoly objects etc. They need to be added as the coefficients of the Lie polynomials. 
        
        # Assemble output Hamiltonians
        out = {}
        lamb0 = self.components[0]
        kx = (lamb0 + lamb0.conjugate())/2 # lamb0.real; .conjugate() works with floats and other objects better than .real etc.
        ky = (lamb0.conjugate() - lamb0)/2/1j # -lamb0.imag
        # N.B.: Hfull = psigma - (1 + kx*x + ky*y)*drift - G
        drift_s = self._prop['drift_sqrt'] # the square root expression in the Hamiltonian belonging to the drift
        H_drift = psigma - drift_s
        H_field = - drift_s*(x*kx + y*ky) - G
        H_full = H_drift + H_field
        
        # remove any remaining constants; they do not affect the equations of motion.
        H_full.pop((0, 0, 0, 0, 0, 0), None)
        H_drift.pop((0, 0, 0, 0, 0, 0), None)
        H_field.pop((0, 0, 0, 0, 0, 0), None)

        out = {}
        out['kx'] = kx
        out['ky'] = ky
        out['full'] = H_full
        out['kick'] = H_field
        out['G'] = G
        out['g'] = g
        self._prop.update(out)
        self.full_hamiltonian = H_full

class multipole(cfm):
    def __init__(self, str=0, n: int=0, *args, **kwargs):
        '''
        Model of a multipole with exactly 2*n pole faces.
        
        Parameters
        ----------
        n: int, optional
            Defines the number 2*n of pole faces.

         
        expansion: int, optional
            An additional parameter (>= n) by which one can control the number of terms in the expansion of the field.
        '''
        if n == 0: # drift case
            components = [0]
        else: # n > 0
            components = [0]*max([2, n]) # max([2, n]) because if n = 1 (the dipole case) then, since (1 + Kx*x + Ky*y)*(By + 1j*Bx) = - \partial_x G + 1j*\partial_y G (see p. 53 top in Ref. [2]), G must be of order <= 2.
            # The real and imaginary part of 'str' are understood to define the normal and skew components of the sextupole.
            # These components are given with respect to the x-axis (tilt 0) and an axis which is rotated relative to x by an angle of pi/2/n.
            #
            # Equivalently, these components are given by the real and imaginary parts of c_n with B_y + i B_x = c_n*(x + iy)**(n - 1), see e.g.
            # Eq. (1.8) in Ref. [2], using y=0 here. The c_n are just the components of the cfm.
            # Therefore:
            components[n - 1] = str
        cfm.__init__(self, components=components, *args, **kwargs)
        
class drift(multipole):
    def __init__(self, *args, **kwargs):
        multipole.__init__(self, *args, n=0, **kwargs)
        
class dipole(multipole):
    def __init__(self, *args, **kwargs):
        multipole.__init__(self, *args, n=1, **kwargs)
        
class quadrupole(multipole):
    def __init__(self, *args, **kwargs):
        multipole.__init__(self, *args, n=2, **kwargs)
        
class sextupole(multipole):
    def __init__(self, *args, **kwargs):
        multipole.__init__(self, *args, n=3, **kwargs)
    
class octupole(multipole):
    def __init__(self, *args, **kwargs):
        multipole.__init__(self, *args, n=4, **kwargs)
        
class rfc(hard_edge_element):
    
    def __init__(self, voltage, phase, frequency, beta0, *args, **kwargs):
        '''
        A generic RF cavity based on the simplest of all models.
                
        Reference(s):
        [1] A. Wolski: Beam Dynamics in High Energy Particle Accelerators.
        '''
        assert 0 < beta0 and beta0 < 1
        self.voltage = voltage
        self.phase = phase
        self.frequency = frequency # Actually, translated to: k = 2*np.pi/constants.speed_of_light*1/T
        self.beta0 = beta0
        hard_edge_element.__init__(self, *args, beta0=beta0, **kwargs)
        
    def calcHamiltonian(self, p=10, **kwargs):
        '''
        The Hamiltonian of a simplified RF cavity., see Ref. [1], Eq. (3.138) p. 112.
        '''
        hard_edge_element.calcHamiltonian(self, **kwargs)
        x, y, sigma, px, py, psigma = self._prop['coords']
        #k = 2*np.pi*self.frequency/constants.speed_of_light # 2.40483/radius # Eq. (3.132) in [1] and earlier: omega = k/c
        #hamiltonian = construct(cos, sigma/beta0*-k + self.phase, **kwargs)*self.voltage
        rf_potential = construct(cos, -sigma/self.beta0*self.frequency + self.phase, power=p, **kwargs)*self.voltage/float(np.pi)
        hamiltonian = self._prop['full'] - rf_potential
        hamiltonian.pop((0, 0, 0, 0, 0, 0), None) # remove any constant term
        self.full_hamiltonian = hamiltonian
        
        
        
    

        


from legacycontour import _cntr as cntr
import numpy as np
import matplotlib.pyplot as plt
from lieops.solver import getRealHamiltonFunction

def plot_contours(rham, xrange, yrange, levels=[], q=[], p=[], **kwargs):
    '''
    Show the contours of a real-valued Hamiltonian.
    
    plane:
        define what components to show
    '''
    # here we do not use matplotlib contourf, instead an older library by which we can also separate the contour
    # from the plotting.
    assert len(q) == len(p)
        
    X, Y = np.meshgrid(xrange, yrange)
    h = rham(X, Y)
    
    if len(q) > 0 and len(p) > 0:
        for k in range(len(q)):
            levels.append(rham(q[k], p[k]))
    assert len(levels) > 0

    c = cntr.Cntr(X, Y, h)
    for level in levels:
        nlist = c.trace(level, level, 0)
        segs = nlist[:len(nlist)//2]

        for seg in segs:
            plt.plot(seg[:,0], seg[:,1], **kwargs)
    return plt


# Collection of some common routines, used to plot and perform standard tasks

def f_identity(*x):
    return x

def f_compose(f, g):
    return lambda *x: f(*g(*x))

def multiturn(bl, xi0, n_reps: int, post=lambda x: x):
    '''
    Perform tracking for a given number of repetitions.
    
    Parameters
    ----------
    bl: beamline, lieoperator or poly
        The object through which we want to track.
    
    n_reps: int
        The number of repetitions.
        
    xi0: The start vector xi0 = [xi0_1, xi0_2, ...]. By default: xi0_k = (q_k + i*p_k)/2, where q_k and
         p_k are the canonical coordinates in ordinary phase space.
    
    post: callable, optional
        An optional function to be applied after bl. 
        Example: We want to compute (A o bl o B)**n at xi0, where A = B**(-1) are maps to normal form.
        Then we need to record for Y0 := B(xi0) the values A(bl(Y0)), A(bl**2(Y0)) etc. Here A = post must
        be inserted.
    
    Returns
    -------
    list
        A list so that the k'th entry corresponds to the k-th component for the requested turns.
    '''
    dim = len(xi0)
    point = xi0
    points = [[] for c in range(dim)]
    for k in range(n_reps):
        point = bl(point)
        post_point = post(point)
        for c in range(dim):
            points[c].append(post_point[c])
    return points


def detuning(nfdict):
    '''
    Get the detuning with amplitude terms for the given normal form dictionary which is returned by the bnf routine.
    
    Parameters
    ----------
    dict
        The output of the bnf routine.
    
    Returns
    -------
    dict
        A dictionary mapping the powers of (xi, eta)-coordinates to the coefficients in the Hamiltonian.
        These powers are equal in the sense that the resulting expression corresponds to a polynomial in
        the actions epsilon_k = xi_k*eta_k.
    '''
    hks = nfdict['Hk']
    hk_last = hks[-1]
    dim = hk_last.dim
    nforder = nfdict['order']
    # nforder: The order of the current normal form: Note that for every order j the 
    # normalized Hamiltonian includes the power (xi*eta)**(j + 1), while the homogeneous polynomials
    # reach up to power j + 2.
    out = {}
    for k, v in hk_last.items():
        k1, k2 = k[:dim], k[dim:]
        if k1 == k2 and sum(k1) <= nforder + 1: 
            # The first condition ensures that the term depends only on the actions.
            # The second condition ensures that we do not include too many powers which may 
            # be beyond the given normal form order.
            out[tuple(k1 + k1)] = v
    return out

def _depsilon(DH, k):
    '''
    Derive a given action-polynomial, represented by a dictionaly-like object,
    with respect to the k-th action. 
    
    Parameters
    ----------
    Hnf: poly or dict
        A polynomial or dict representing the powers depending only on its actions xi_k*eta_k.
        
    k: integer
        Index with respect to which we want to derive.
    '''
    dim = len(next(iter(DH.keys())))//2
    assert k <= dim - 1, 'Requested index of derivative direction larger than given dimension.'
    D = {}
    for j, v in DH.items():
        j1, j2 = j[:dim], j[dim:]
        assert j1 == j1, 'Given polynomial does not purely depend on its actions.'
        if j1[k] == 0:
            continue
        new = [jm for jm in j1]
        new[k] = j1[k] - 1
        D[tuple(new + new)] = v*j1[k]
    return D

def getPhaseAdvanceFunc(*deps):
    '''
    Create a function giving the phase advance with respect to the action(s).
    
    Parameters
    ----------
    deps
        The output of the _depsilon routine for each direction.
        Example: deps = *[_depsilon(detuning_dict, k) for k in range(dim)]

    Returns
    -------
    callable
        A function taking values in *actions and s, returning the phase advance according to (e.g.)
        Ref. [1], Eq. (1.47b).
        
    Reference(s):
    [1]: M. Titze: "Space Charge Modeling at the Integer Resonance for the CERN PS and SPS", PhD Thesis (2019).
    '''
    dim = len(deps)
    def pa(*action, s: float=1):
        '''
        Compute the phase advance in normal form.
        
        Parameters
        ----------
        *action
            The action(s) at which to obtain the phase advance.
            
        s: float, optional
            The free parameter defining the phase advance.
            
        Returns
        -------
        list
            A list where the k'th entry corresponds to the detuning of the k'th phase advance.
        '''
        # Example for dim = 2: deps[k].keys() = (0, 1, 0, 1), (1, 2, 1, 2) , ... for k = 0
        out = []
        for k in range(dim):
            sumk = 0
            for p, v in deps[k].items():
                prod = 1
                for j in range(dim):
                    if p[j] == 0: # Ensures that 0**0 = 1 so that e.g. (0, 0) (the constant term) is included
                        continue
                    prod *= action[j]**p[j]
                sumk += prod*v
            out.append(sumk*s)
        return out
    return pa



from functools import reduce
import numpy as np

from accphys.tools import _depsilon, detuning
from lieops import poly, lexp, create_coords

from .tools import f_identity, f_compose

class nf:
    '''
    Class to handle operations related to the normal form of a given Hamiltonian.
    
    Reference(s):
    [1]: M. Titze, "Space Charge Modeling at the Integer Resonance for the CERN PS and SPS", PhD Thesis (2019).
    '''
    def __init__(self, hamiltonian, **kwargs):
        self.hamiltonian = hamiltonian
        self.dim = hamiltonian.dim
        self.setup_nf(**kwargs)
        
    def setup_nf(self, t_nf=-1, **kwargs):
        '''
        Calculate the Birkhoff normal form of the given Hamiltonian, up to a specific order.
        
        Parameters
        ----------
        t_nf: float, optional
            The length of the integration step for the normal form Hamiltonian.
        
        **kwargs
            Optional arguments passed to self.hamiltonian.bnf routine.
        '''
        # power should be sufficiently high; self.lo_power = 10 is not sufficient for nfRot. Better 20.
        # TODO:
        # - may use self.lo_power instead.
        # - replace steps by symplectic integrator to prevent the use of max_power etc.
        # - Only the action on xi may be necessary, but currently I keep the full range to avoid any unecessary bugs.
        self._lo_power = 20
        self.t_nf = t_nf
        
        self.__dict__.update(self.hamiltonian.bnf(**kwargs))
        self.detuning = detuning(self.__dict__)
        self.dHdaction = [poly(values=_depsilon(self.detuning, k), dim=self.dim) for k in range(self.dim)]
        # we use poly objects here, because in the next line we multiply these polynomials with other terms.
        # Also, we want to apply the resulting function on coordinates.
        
        self._setup_nfRot(t_nf=t_nf)
        self._setup_chimaps()
        self.transform(**kwargs)
        
    def _setup_nfRot(self, **kwargs):
        self.t_nf = kwargs.get('t_nf', self.t_nf)
        _xieta = create_coords(dim=self.dim, max_power=self.hamiltonian.max_power)
        self._lHnfmap = lexp(poly(values=self.detuning, dim=self.dim), t=self.t_nf, power=self._lo_power)(*_xieta)
        self.nfRot = lambda *z: [lhfm(*z) for lhfm in self._lHnfmap] # N.B. in general it is important that both xi and eta components are returned here.
        # alternative (requires exp with multiplication (not ad)):
        #self.nfRot = lambda *xieta: [exp(self.dHdaction[k]*1j, power=self.lo_power)*xieta[k] for k in range(self.dim)] +\
        #                            [exp(self.dHdaction[k]*-1j, power=self.lo_power)*xieta[k + self.dim] for k in range(self.dim)] 
        # (1.48) in Ref. [1] for the case of a single 'a' or 'b'-value (i.e. a_k = 1)
        
    def _setup_chimaps(self):
        _xieta = create_coords(dim=self.dim, max_power=self.hamiltonian.max_power)
        self._lchi = [lexp(chi, t=1, power=self._lo_power) for chi in self.chi] # the Lie-operators belonging to the chi-transformations
        self._lchi_inv = [lexp(chi, t=-1, power=self._lo_power) for chi in self.chi] # the inverse Lie-operators belonging to the chi-transformations
        
        self._Amap = reduce(f_compose, self._lchi[::-1], f_identity)(*_xieta) # the map from ordinary (xi, eta)-space to normal-form space 
        self._A_invmap = reduce(f_compose, self._lchi_inv, f_identity)(*_xieta) # the map from normal-form space to ordinary (xi, eta)-space
        self.A = lambda *xieta: [A(*xieta) for A in self._Amap]  # N.B. in general it is important that both xi and eta components are returned here.
        self.A_inv = lambda *xieta: [Ai(*xieta) for Ai in self._A_invmap]  # N.B. in general it is important that both xi and eta components are returned here.

    def transform(self, kind='cnf', **kwargs):
        '''
        Select a linear transformation which should be executed when tracking before and after the non-linear maps to NF.
        
        Parameters
        ----------
        kind: str, optional
            A string denoting the nature of the transformation. Supported type(s) are:
            'cnf': Complex normal form
            'ops': Ordinary phase space
        '''
        supported_types = ['cnf', 'ops']
        self._transform_kind = kind
        if kind == 'cnf':
            # Complex normal form
            self.setTransform(np.eye(self.dim*2))
        elif kind == 'ops':
            # Ordinary phase space
            self.setTransform(self.nfdict['K'], self.nfdict['Kinv'])
        else:
            raise RuntimeError(f"'{kind}' not recognized. Supported type(s): {supported_types}")
        
    def setTransform(self, N, N_inv=[]):
        '''
        Set linear transformation maps which should be executed when tracking before and after the non-linear maps to NF.
        
        Parameters
        ----------
        N: array_like
            The linear map to be applied.
        
        N_inv: array_like, optional
            The linear map to be applied after the operations. If nothing specified, the inverse of N will be used. 
        '''
        assert N.shape == (self.dim*2, self.dim*2)
        if len(N_inv) == 0:
            N_inv = np.linalg.inv(N)
        self._transform_A2 = lambda *z: [sum([z[l]*N[k, l] for l in range(self.dim*2)]) for k in range(self.dim*2)] # we use matrix multiplication by hand, to avoid numpy complains with nested arrays in case z[l] are numpy arrays; N[k, l] stands on the right here, to avoid possible numpy issues with right-multiplications
        self._transform_A2_inv = lambda *z: [sum([z[l]*N_inv[k, l] for l in range(self.dim*2)]) for k in range(self.dim*2)]
        
    def multiturn(self, *xieta, n_reps=1, **kwargs):
        # Preparation steps
        # 1. Determine requested transformation
        kind_inp = kwargs.get('kind', 'cnf') # if nothing specified, the default option will be 'cnf'
        kind = getattr(self, '_transform_kind', kind_inp)
        if kind != kind_inp:
            self.transform(kind=kind_inp)
        # 2. Determine requested integration length    
        t_nf = kwargs.get('t_nf', self.t_nf)
        if self.t_nf != t_nf:
            # re-calculate inner NF rotation with the new length
            self._setup_nfRot(t_nf=t_nf)
        
        # Tracking
        points = [self._transform_A2_inv(*self._transform_A2(*xieta))] # Applying A2 and back again ensures that the output will be in a common format (e.g. if one component is given in terms of a numpy array and the other component a float etc.
        xieta = self._transform_A2(*xieta) # apply the linear map first
        xi_nf = self.A(*xieta)
        for k in range(n_reps):
            xi_nf = self.nfRot(*xi_nf)
            points.append(self._transform_A2_inv(*self.A_inv(*xi_nf)))
        return points
        
    def oneTurnMap(self, *xieta, **kwargs):
        kwargs['n_reps'] = 1
        return self.multiturn(*xieta, **kwargs)[-1]
    
    def __call__(self, *xieta, **kwargs):
        return self.oneTurnMap(*xieta, **kwargs)
    


from functools import reduce
from tqdm.auto import tqdm

from lieops import create_coords, combine
from .tools import f_identity, f_compose

class beamline:
    
    def __init__(self, *elements, **kwargs):
        '''
        Class to model an accelerator beamline.
        '''
        # consistency checks
        assert len(elements) > 0
        assert all([hasattr(e, 'hamiltonian') for e in elements])
        dim0 = elements[0].hamiltonian.dim
        assert all([e.hamiltonian.dim == dim0 for e in elements]), 'Dimensions of the individual Hamiltonians differ.'
        
        self.elements = [e.copy() for e in elements]
        self.ordering = kwargs.get('ordering', list(range(len(elements))))
        
    def __len__(self):
        return len(self.ordering)
    
    def __getitem__(self, key):
        return self.elements[self.ordering[key]]
    
    def __setitem__(self, key, value):
        if value not in self:
            self.ordering[key] = len(self.elements)
            self.elements.append(value)
        else:
            index = self.elements.index(value)
            self.ordering[key] = index
            
    def index(self, value):
        return self.elements.index(value)
    
    def append(self, value):
        if value not in self.elements:
            index = len(self.elements)
            self.elements.append(value)
        else:
            index = self.elements.index(value)
        self.ordering.append(index)
        
    def lengths(self):
        '''
        Return the lengths of the individual elements.
        '''
        return [e.length for e in self]
    
    def get_dim(self):
        return self[0].hamiltonian.dim
        
    def setHamiltonians(self, *args, **kwargs):
        '''
        Project the Hamiltonians of the individual elements to specific dimensions.
        '''
        for e in self.elements:
            e.setHamiltonian(*args, **kwargs)
            
    def magnus(self, power=1, **kwargs):
        '''
        Combine the individual Hamiltonians by means of Magnus series.
        
        Parameters
        ----------
        power: int, optional
            Power in terms of the free parameter (usually the longitudinal coordinate s),
            up to which the magnus series will be exact. 
        
        **kwargs
            Optional keyword arguments passed to lieops.ops.lie.combine routine
        '''            
        hamiltonians = [e.hamiltonian for e in self]
        self._magnus_series, self._magnus_hamiltonian, self._magnus_forest = combine(*hamiltonians, power=power, 
                                                                                     lengths=[e.length for e in self], **kwargs)
        return sum(self._magnus_series.values())
        
    def calcFlows(self, t=-1, **kwargs):
        '''
        Calculate the sequence of Lie-operators exp(:f_n:) for each element f_n in self.sequence.
        
        Parameters
        ----------
        Parameters
        ----------
        lengths: list, optional
            Lengths of the elements. These lengths will modifying the respecive flows accordingly.
            If given, then exp(lengths[k]:x[k]:) will be computed. An additional t-argument will
            compute exp(t*lengths[k]:x[k]:).

        **kwargs
            Optional parameters passed to lieops.ops.lie.poly.flow
        '''
        element_flows = [e.hamiltonian.flow(t=t*e.length, **kwargs) for e in self.elements] # compute the flows of the unique elements
        self.flows = [element_flows[j] for j in self.ordering]
        
    def calcOneTurnMap(self, half=False, **kwargs):
        '''
        Parameters
        ----------
        half: boolean, optional
            If specified, only compute the result for the xi-polynomials.
            
        **kwargs 
            Optional parameters passed to create_coords routine (e.g. any possible max_power)
        '''
        assert hasattr(self, 'flows'), 'Need to call self.calcFlows first.'
        dim0 = self.get_dim()
        xiv = create_coords(dim0, **kwargs)
        if half:
            xiv = xiv[:dim0]
        # N.B. 'reduce' will apply the rightmost function in the given list first, so that e.g.
        # [f0, f1, f2]
        # will be executed as
        # f0(f1(f2(z)))
        # etc.
        # Since in our beamline the first element in the list should be executed first,
        # we have to reverse the order here.
        composition = reduce(f_compose, self.flows[::-1], f_identity)
        self.oneTurnMap = composition(*xiv)
        
    def __call__(self, *point):
        assert hasattr(self, 'oneTurnMap'), 'Need to call self.calcOneTurnMap first.'
        return [c(*point) for c in self.oneTurnMap]
    
    def multiturn(self, *xi0, n_reps: int=1, post=lambda x: x):
        '''
        Perform tracking for a given number of repetitions.

        Parameters
        ----------
        n_reps: int
            The number of repetitions.

        xi0: The start vector xi0.

        post: callable, optional
            An optional function to be applied after bl. 
            Example: We want to compute (A o bl o B)**n at xi0, where A = B**(-1) are maps to normal form.
            Then we need to record for Y0 := B(xi0) the values A(bl(Y0)), A(bl**2(Y0)) etc. Here A = post must
            be inserted.

        Returns
        -------
        list
            A list so that the k'th entry corresponds to the k-th component for the requested turns.
        '''
        point = xi0
        points = []
        for k in range(n_reps):
            point = self(*point)
            post_point = post(point)
            points.append(post_point)
        return points

        
        
        
        


from .beamline import beamline
from .elements import cfm
from tqdm.auto import tqdm

from latticeadaptor.core import LatticeAdaptor
import numpy as np
from njet.jet import factorials

def prepare_df(hdf, position_label='s', length_label='L', position=0, tol=1e-6, **kwargs):
    '''
    Prepare a given data frame so that there will be no zero-spaces in between two elements.
    
    Parameters
    ----------
    hdf: pd.Dataframe
        Pandas dataframe object defining the beam sequence. The dataframe must contain the
        position_label and length_label columns.
        
    position_label: str, optional
        Label denoting the position of the individual elements within the dataframe.
        
    length_label: str, optional
        Label denoting the lengths of the individual elements within the dataframe.
    
    position: float, optional
        Defines the alignment of the elements relative to their actual position.
        Take a value between 0 and 1, where 0 means the position is given relative to the element start,
        while 1 means that they are taken with respect to the element end.
    
    tol: float, optional
        Tolerance below which we consider the end/start of two adjacent elements to agree with each other.
        
    Returns
    -------
    Pandas dataframe
        A dataframe object in which all elements are positioned so that there are no gaps inbetween (drift
        elements are added accordingly).
    '''
    
    toStartPos = lambda z, length: z - length*position # transform position to start of element.
    
    # ensure the Dataframe index is reset
    hdf = hdf.reset_index(drop=True)
    
    column_labels = list(hdf.columns)
    position_label_index = column_labels.index(position_label)
    length_label_index = column_labels.index(length_label)
    
    # Iterate through the entire sequence to check the position and lengths; insert drifts if necessary
    for k in range(len(hdf)):
        current_row = hdf.iloc[k]
        pos_1 = current_row[position_label]
        length = current_row[length_label]
        pos = toStartPos(pos_1, length)
        
        if k + 1 < len(hdf):
            next_row = hdf.iloc[k + 1]
            next_pos_1 = next_row[position_label]
            next_length = next_row[length_label]
            next_pos = toStartPos(next_pos_1, next_length)
        else:
            next_pos = pos + length

        if pos + length > next_pos + tol:
            # in this case the next element is overlapping the previous element due to a too large length.
            raise RuntimeError(f"Element at row-index {k}, covering [{pos}, {pos + length}] appears to overlap its successor, beginning at {next_pos}. Check lattice input and/or adjust position: {position} and/or tol: {tol}.")

        if pos + length + tol < next_pos:
            # here we have to insert additional drift space in between.
            empty_space = next_pos - pos - length
            
            new_row = [0]*len(hdf.columns)
            new_row[position_label_index] = pos + length
            new_row[length_label_index] = empty_space
            hdf.loc[k + 0.5] = new_row # insert a new row in between k and k + 1
            
    return hdf.sort_index().reset_index(drop=True)


def to_beamline(hdf, beta0, component_labels, position_label='s', length_label='L', **kwargs):
    '''
    Construct a beamline from a given lattice.
    
    Parameters
    ----------
    hdf: Pandas dataframe
        A Pandas dataframe object containing the position, lengths and field strengths of the individual
        elements in the beamline.
        
    position_label: str, optional
        Label denoting the position of the individual elements within the dataframe.
        
    length_label: str, optional
        Label denoting the lengths of the individual elements within the dataframe.
        
    **kwargs
        Keyworded arguments passed to the cfm element.
        
    Returns
    -------
    beamline
        A beamline object representing the sequence of elements.
    '''
    # Preparation; ensure that no empty space exists between elements (they will be filled with drifts if necesary):
    hdf = prepare_df(hdf, position_label=position_label, length_label=length_label, **kwargs)
            
    # group the given elements with respect to the remaining labels & find the uniques among them
    group_labels = [c for c in hdf.columns if c != position_label] # labels by which we distinguish different elements; custom columns may be added to artifically distinguish different element groups
    group_index_label = 'group_index'
    grp = hdf.groupby(group_labels, sort=False)
    hdf[group_index_label] = grp.ngroup()
    unique_elements = hdf.drop_duplicates(group_index_label)
    
    # build the elements:
    elements = []
    group_index = 0
    for n in tqdm(range(len(unique_elements))):
        row = unique_elements.iloc[n]
        components = [row[c] for c in component_labels]
        length = row[length_label]
        assert group_index == row[group_index_label] # verify that the position in the element list corresponds with the group index given by ngroup.
        elements.append(cfm(beta0=beta0, components=components, length=length, **kwargs))
        group_index += 1
        
    # set the ordering
    ordering = list(hdf[group_index_label])
            
    return beamline(*elements, ordering=ordering)


def from_madx(filename, beta0, **kwargs):
    '''
    Load MAD-X lattice from file, using LatticeAdaptor module, 
    and construct a beamline object from the data.
    
    Parameters
    ----------
    filename: str
        The name of the MAD-X lattice to be loaded.
        
    beta0: float
        The realtivistic beta-factor (related to the energy of the beam). This is required later to
        build the Hamiltonians.
        
    **kwargs
        Optional arguments passed to 'to_beamline' routine.
        
    Returns
    -------
    beamline
        A beamline object representing the sequence of elements in the given lattice.
        
    Pandas dataframe
        A Pandas dataframe object, representing the loaded sequence of the lattice.
    '''
    la = LatticeAdaptor()
    la.load_from_file(filename, ftype='madx')
    raw_df = la.table
    
    # MAD-X specific labels
    position_label='at'
    length_label='L'
    bend_kx_label = 'K0'
    angle_label = 'ANGLE'
    component_labels = [bend_kx_label, 'K1', 'K2']
    madx_default_position = 0.5 # MAD-X tends to denote the position of the elements in the center

    if bend_kx_label not in raw_df.columns:
        # add kx
        angles = raw_df[angle_label].values
        lengths = raw_df[length_label].values 
        valid_indices = np.logical_and((~np.isnan(angles)), lengths > 0)
        kx = np.zeros(len(raw_df))
        kx[valid_indices] = angles[valid_indices]/lengths[valid_indices] # r*phi = L; kx = 1/r
        raw_df[bend_kx_label] = kx
    
    # drop elements with zero length and uneccesary columns;
    # N.B. E1 and E2 denote rotation angles of the pole-faces. If they are non-zero,
    # they are usually half the bend angle (in the rectangular case). We will ignore rectangular
    # bends for the time being and use s-bends here.
    columns_oi = [position_label, length_label] + component_labels
    # if they exist, add skew-values to the components; TODO: check & verify this numerically
    for j in range(1, 3):
        if f'K{j}S' in raw_df.columns:
            raw_df[f'K{j}'] = raw_df[f'K{j}'].values + raw_df[f'K{j}S']*1j
    raw_df = raw_df.loc[raw_df[length_label] > 0][columns_oi]
    
    # (TO BE CHECKED; TODO)
    facts = factorials(3)
    for j in range(1, 3):
        raw_df[f'K{j}'] = raw_df[f'K{j}'].values/facts[j]
        
    # construct the sequence of Lie operators
    kwargs['position'] = kwargs.get('position', madx_default_position) 
    seq2 = to_beamline(raw_df, beta0=beta0, 
                      component_labels=component_labels, position_label=position_label,
                      length_label=length_label, **kwargs)
    return seq2, raw_df


import heyoka as hy

from lieops.solver import realHamiltonEqs
from lieops.solver.heyoka import createHamiltonEqs


def create_full_hy_hamiltonian(bl, tol=1e-15):
    '''
    The heyoka solver can only deal with Hamiltonians which depend smoothly on the time-parameter t.
    
    For hard-edge elements we have to introduce auxiliary parameters which will switch off/on
    the Hamiltonian of a specific element after a certain time. These parameters are introduced here.
    '''
    qeq, peq = 0, 0
    for k in range(len(bl)):
        qp, hameqs = createHamiltonEqs(bl[k].hamiltonian, tol=tol)
        qeq += hameqs[0]*hy.par[k]
        peq += hameqs[1]*hy.par[k]
    return [e for e in zip(*[qp, [qeq] + [peq]])]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['accphys',
 'accphys..ipynb_checkpoints',
 'accphys.solver',
 'accphys.solver..ipynb_checkpoints']

package_data = \
{'': ['*']}

install_requires = \
['ipykernel>=5.5.5,<6.0.0',
 'latticeadaptor>=0.3.9',
 'lieops>=0.1.1,<0.2.0',
 'matplotlib>=3.5.1,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'scipy>=1.8.0,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

setup_kwargs = {
    'name': 'accphys',
    'version': '0.1.0',
    'description': 'Analyze non-linear single-particle storage ring dynamics',
    'long_description': None,
    'author': 'Malte Titze',
    'author_email': 'mtitze@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
